;;; fsap.el ---find symbol at point

;; Copyright (C) 2008 Martin Rudalics

;; Time-stamp: "2009-03-01 20:34:21 martin"
;; Author: Martin Rudalics <rudalics@gmx.at>
;; Keywords: tags, etags

;; GNU Emacs is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 3, or (at your option)
;; any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
;; Boston, MA 02110-1301, USA.

(defvar fsap-name nil)

(defvar fsap-symbol nil)

(defvar fsap-list nil)

(defvar fsap-deque nil)

(defsubst fsap-rotate ()
  (setq fsap-deque (nconc (cdr fsap-deque) (list (car fsap-deque)))))

(defsubst fsap-reject ()
  (setq fsap-deque (cdr fsap-deque))
  nil)

(defun fsap-symbol ()
  "Return Elisp symbol at `point'.
Return nil when there's no such symbol."
  (if (eq (syntax-table) emacs-lisp-mode-syntax-table)
      (let ((name (find-tag-default)))
	(when name (intern-soft name)))
    (with-syntax-table emacs-lisp-mode-syntax-table
      (modify-syntax-entry ?\{ "()  ")
      (modify-syntax-entry ?\} ")(  ")
      (let ((name (find-tag-default)))
	(when name (intern-soft name))))))

(defun fsap-find-symbol (symbol)
  (let (value)
    (while (and fsap-deque (not value))
      (or (and (eq (car fsap-deque) 'defun)
	       (or (and (fboundp symbol)
			(setq value (find-definition-noselect symbol nil))
			(fsap-rotate))
		   (fsap-reject)))
	  (and (eq (car fsap-deque) 'defvar)
	       (or (and (or (boundp symbol)
			    (get symbol 'variable-documentation))
			(setq value (find-definition-noselect symbol 'defvar))
			(fsap-rotate))
		   (fsap-reject)))
	  (and (eq (car fsap-deque) 'defface)
	       (or (and (facep symbol)
			(setq value (find-definition-noselect symbol 'defface))
			(fsap-rotate))
		   (fsap-reject)))))
    value))

(defun fsap-find (&optional other)
  "Go to symbol or tag around `point'."
  (interactive)
  ;; Set up `fsap-deque'.
  (let ((repeat (memq last-command '(fsap-find fsap-back))))
    (unless repeat
      (setq fsap-deque '(defun defvar defface)))
    ;; Try our methods.
    (let ((name (and repeat fsap-name))
	  (symbol (or (and repeat fsap-symbol)
		      (fsap-symbol)))
	  (buffer (current-buffer))
	  (point (point))
	  tags-case-fold-search
	  name value)
      ;; Try to find symbol or tag.
      (or (and symbol
	       (setq value (fsap-find-symbol symbol))
	       (setq fsap-symbol symbol)
	       (setq fsap-name (symbol-name symbol)))
	  (and (eq last-command this-command)
	       ;; Repeat search.
	       (setq value (find-tag-noselect fsap-name t)))
	  (and (setq name (or name (find-tag-default)))
	       (file-readable-p
		(setq tags-file-name
		      (concat
		       (expand-file-name default-directory) "TAGS")))
	       (setq value (find-tag-noselect name))
	       (progn
		 (setq fsap-symbol nil)
		 (setq fsap-name name)))
	  (and (setq symbol (car-safe (find-function-read)))
	       (fboundp symbol)
	       (setq value (find-definition-noselect symbol nil))
	       (setq fsap-symbol symbol)
	       (setq fsap-name (symbol-name symbol))))

      (when value
	(let ((buffer (or (car-safe value) value))
	      (point (cdr-safe value)))
	  (if other
	      (pop-to-buffer buffer t)
	    (switch-to-buffer buffer))
	  (when point (goto-char point)))
	;; Update `fsap-list'.
	(let ((parameter (list buffer point fsap-symbol fsap-name)))
	  (set-window-parameter
	   nil 'fsap-list
	   (cons parameter (window-parameter nil 'fsap-list))))))))

(defun fsap-find-other ()
  "Go to symbol or tag around `point' in another window."
  (interactive)
  (fsap-find t))

(defun fsap-back ()
  "Go back to last position after `fsap-find'."
  (interactive)
  (let ((parameter (car-safe (window-parameter nil 'fsap-list))))
    (when parameter
      (set-window-buffer (selected-window) (nth 0 parameter))
      (goto-char (nth 1 parameter))
      (setq fsap-symbol (nth 2 parameter))
      (setq fsap-name (nth 3 parameter))
      (set-window-parameter
       nil 'fsap-list
       (cdr (window-parameter nil 'fsap-list))))))

(global-set-key [(meta return)] 'fsap-find)
(global-set-key [(meta backspace)] 'fsap-back)
(global-set-key [(shift meta return)] 'fsap-find-other)

(setq fsap-list nil)
